// FTPClient.java
// FTP client implementation class

package JP.gr.java_conf.tech.ftp;

import java.io.*;
import java.net.*;
import java.util.StringTokenizer;
import java.util.Vector;
import JP.gr.java_conf.tech.net.Session;

/**
	FTPNCAgNX

	@version	1.2.1	09/17/1999
	@author		Takahiro Shigemoto
*/
public class FTPClient extends Session
{
	/* FTPڑ|[g\킷萔 */
	public static final int FTP_PORT = 21;

	// FTPT[o^[R[hꗗ
	static final int OPEN = 150;
	static final int COMMAND = 200;
	static final int HELP = 214;
	static final int NAME = 215;
	static final int READY = 220;
	static final int QUIT = 221;
	static final int TRANSFER = 226;
	static final int PASS = 230;
	static final int FILE = 250;
	static final int DIRECTORY = 257;
	static final int USER = 331;
	static final int FILE_OTHER = 350;

	// f[^Mp\Pbg|[g
	ServerSocket ssocket;

	// M^Cv
	private static final int TYPE_ASCII  = 0;
	private static final int TYPE_BINARY = 1;

	// NCAgIPAhX
	private int local_ip[] = new int[4];

	// NCAg̉sR[h
	private final String line_separator = System.getProperty("line.separator");

	// T[õt@CZp[^
	private String server_file_separator = "/";

	// FTPT[ȍ
	private String ftp_server_info;

	// GR[h
	private String default_encoding = System.getProperty("file.encoding");

	// 擾ItZbg(ʒu)
	private int offset = 0;

	/**
		T[oA[U[ApX[hw肵FTPT[o֐ڑ
	*/
	public FTPClient(String server_name,String user_name,String password)
	{
		// T[oւ̐ڑ
		super(server_name, FTP_PORT);
		this.ftp_server_info = recvData();
		if(!checkReturnCode(ftp_server_info, READY))
			ErrorMessage("Server isn't ready");

		checkReturnCodeLoop(READY);

		// [U[F
		Authentication(user_name,password);

		// T[õVXe̎擾
		String server_info = sendSystemCommand();

		// VXeAt@CZp[^̎擾
		if(server_info.indexOf("UNIX") != -1)
			server_file_separator = "/";
		else
			server_file_separator = "/";

		// NCAgIPAhX擾
		try
		{
			String str_local_ip = InetAddress.getLocalHost().getHostAddress();
			StringTokenizer st = new StringTokenizer(str_local_ip, ".");

			for(int i=0; st.hasMoreTokens(); i++)
				local_ip[i] = Integer.parseInt(st.nextToken());
		}
		catch(UnknownHostException uh_ex)
		{
			uh_ex.printStackTrace();
			ErrorMessage("Can't get Client IP Adderss");
		}
	}

	/**
		FTPT[o̎擾
		@return	T[o
	*/
	public String getServerInfo()
	{
		int begin_index = ftp_server_info.indexOf(" ");
		int end_index = ftp_server_info.lastIndexOf(" ");

		return(ftp_server_info.substring(begin_index + 1, end_index));
	}

	/**
		[U[F؂s
		@param	user_name	[U[
		@param	password	pX[h
	*/
	private void Authentication(String user_name,String password)
	{
		sendData("USER " + user_name);
		if(!checkReturnCode(USER))
			ErrorMessage("UserName is Bad");

		sendData("PASS " + password);
		if(!checkReturnCode(PASS))
			ErrorMessage("Password is Bad");

		checkReturnCodeLoop(PASS);
	}

	/**
		M^CvI
		@param	type	TYPE_ASCII,TYPE_BINARŶꂩ
	*/
	private void selectType(int type)
	{
		if(type == TYPE_ASCII)
			sendData("TYPE A");
		else if(type == TYPE_BINARY)
			sendData("TYPE I");

		if(!checkReturnCode(COMMAND))
			ErrorMessage("Select Type Failed");
	}

	/**
		f[^MpNCAg|[g̃I[vƃT[oւ̒ʒm
	*/
	private void openClientPort()
	{
		// f[^Mp\Pbg̏
		try
		{
			// 󂢂Ă|[gőҋ@
			ssocket = new ServerSocket(0);
		}
		catch(IOException io_ex)
		{
			io_ex.printStackTrace();
			ErrorMessage("Can't Initialize Data Receive Socket");
		}

		// T[oւ̃|[g̒ʒm
		int port = ssocket.getLocalPort();
		int high_port_number = port / 256;
		int low_port_number  = port % 256;
		sendData("PORT " + local_ip[0] + "," + local_ip[1] + "," + local_ip[2] + "," + local_ip[3] +
							 "," + high_port_number + "," + low_port_number);

		if(!checkReturnCode(COMMAND))
			ErrorMessage("Can't open Client Port");
	}

	/**
		t@C擾̏(M)
		@param	file_name	ΏۂƂȂt@C
		@return	Ώۃt@C̃t@CTCY
	*/
	private int prepareGetFile(String file_name)
	{
		// NCAg̏
		openClientPort();

		// ItZbg̐ݒ
		if(offset != 0)
			setOffsetCommand();

		// t@C擾R}h̔s
		sendData("RETR " + file_name);
		String receive = recvData();
		if(!checkReturnCode(receive, OPEN))
			ErrorMessage("Get File Failed");

		// T[ỏAt@CTCYo
		int file_size = getFileSizeFromMessage(receive);

		return(file_size);
	}

	/**
		T[obZ[Wt@CTCY擾(M)
		@param	server_message	T[õbZ[W
		@return	bZ[W蒊oꂽt@CTCY
	*/
	private int getFileSizeFromMessage(String server_message)
	{
		int size,begin_index,end_index;

		begin_index = server_message.lastIndexOf('(') + 1;
		end_index   = server_message.lastIndexOf("bytes") - 1;
		size = Integer.parseInt(server_message.substring(begin_index,end_index));

		return(size);
	}

	/**
		wt@C̃t@CTCY擾
		@param	file_name	t@C
		@return	t@CTCY
	*/
	public int getFileSize(String file_name)
	{
		int file_size = prepareGetFile(file_name);
		if(!checkReturnCode(TRANSFER))
			ErrorMessage("Transfer Failed.");

		return(file_size);
	}

	/**
		AXL[t@C(eLXgt@C)̎擾
		@param	file_name	擾t@C
		@return	t@C̓e
	*/
	public String getAsciiFile(String file_name)
	{
		return(getAsciiFile(file_name, default_encoding));
	}

	/**
		AXL[t@C(eLXgt@C)̎擾
		@param	file_name	擾t@C
		@param	kanji_encoding	{R[h(Shift_JIS, EUC-JP, ISO-2022-JP...)
		@return	t@C̓e
	*/
	public String getAsciiFile(String file_name, String kanji_encoding)
	{
		StringBuffer buffer = new StringBuffer();

		// AXL[[hɃZbg
		selectType(TYPE_ASCII);

		// t@C擾̏
		prepareGetFile(file_name);

		try
		{
			// T[o̐ڑ̎󂯕t
			Socket socket = ssocket.accept();

			// AXL[f[^̎擾
			BufferedReader reader = new BufferedReader(new InputStreamReader(socket.getInputStream(), kanji_encoding));
			String buf;
			while((buf = reader.readLine()) != null)
			{
				buffer.append(buf + line_separator);
			}
			reader.close();

			if(!checkReturnCode(TRANSFER))
				ErrorMessage("File Transfer Failed");
		}
		catch(UnsupportedEncodingException ue_ex)
		{
			ue_ex.printStackTrace();
			ErrorMessage(kanji_encoding + " is not supported.");
		}
		catch(IOException io_ex)
		{
			io_ex.printStackTrace();
			ErrorMessage("File IO Error");
		}

		return(buffer.toString());
	}

	/**
		oCit@C̎擾
		@param	buffer		f[^󂯎obt@
		@param	file_name	擾t@C
		@return	Mf[^TCY
	*/
	public int getBinaryFile(byte[] buffer, String file_name)
	{
		return(getBinaryFile(buffer, file_name, buffer.length));
	}

	/**
		oCit@C̎擾
		@param	buffer		f[^󂯎obt@
		@param	file_name	擾t@C
		@param	length		擾f[^TCY
		@return	Mf[^TCY
	*/
	public int getBinaryFile(byte[] buffer, String file_name, int length)
	{
		// oCi[hɃZbg
		selectType(TYPE_BINARY);

		// t@C擾̏
		int file_size = prepareGetFile(file_name);

		if(length > (buffer.length - offset))
			length = (buffer.length - offset);

		try
		{
			// T[o̐ڑ̎󂯕t
			Socket socket = ssocket.accept();
			BufferedInputStream  input  = new BufferedInputStream(socket.getInputStream());

			// oCif[^̎擾
			int read_length = 0;
			while(read_length < length)
			{
				read_length += input.read(buffer, (offset + read_length), (length - read_length));
			}
			input.close();

			if(!checkReturnCode(TRANSFER))
				ErrorMessage("File Transfer Failed");

			return(read_length);
		}
		catch(IOException io_ex)
		{
			io_ex.printStackTrace();
			ErrorMessage("File IO Error");

			return(-1);
		}
	}

	/**
		AXL[t@C̃Abv[h
		@param	file_name	Mt@C
	*/
	public void putAsciiFile(String file_name)
	{
		putAsciiFile(file_name, default_encoding);
	}

	/**
		AXL[t@C̃Abv[h
		@param	file_name	Mt@C
		@param	kanji_encoding	{R[h(Shift_JIS, EUC-JP, ISO-2022-JP...)
	*/
	public void putAsciiFile(String file_name, String kanji_encoding)
	{
		// AXL[[hɃZbg
		selectType(TYPE_ASCII);

		// NCAg̏
		openClientPort();

		// ItZbg̐ݒ
		if(offset != 0)
			setOffsetCommand();

		// t@CMR}h̔s
		sendData("STOR " + file_name);
		if(!checkReturnCode(OPEN))
			ErrorMessage("Get File Failed");

		try
		{
			// Mt@C̎擾
			StringBuffer src = new StringBuffer();
			BufferedReader reader = new BufferedReader(new FileReader(file_name));

			String buf;
			while((buf = reader.readLine()) != null)
			{
				src.append(buf + line_separator);
			}

			// Xg[
			reader.close();

			// T[o̐ڑ̎󂯕t
			Socket socket = ssocket.accept();

			// AXL[t@C̑M
			BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(socket.getOutputStream(),kanji_encoding));
			writer.write(src.toString(), 0, src.length());

			// Xg[
			writer.close();
		}
		catch(IOException io_ex)
		{
			io_ex.printStackTrace();
			ErrorMessage("File IO Error");
		}

		if(!checkReturnCode(TRANSFER))
			ErrorMessage("File Transfer Failed");
	}

	/**
		oCit@C̃Abv[h
		@param	file_name	Mt@C
	*/
	public void putBinaryFile(String file_name)
	{
		// oCi[hɃZbg
		selectType(TYPE_BINARY);

		// NCAg̏
		openClientPort();

		// ItZbg̐ݒ
		if(offset != 0)
			setOffsetCommand();

		// t@CMR}h̔s
		sendData("STOR " + file_name);
		if(!checkReturnCode(OPEN))
			ErrorMessage("Get File Failed");

		try
		{
			// Mt@C̎擾
			File src_file = new File(file_name);
			BufferedInputStream input = new BufferedInputStream(new FileInputStream(src_file));

			byte src[] = new byte[(int)src_file.length()];
			input.read(src);

			// Xg[
			input.close();

			// T[o̐ڑ̎󂯕t
			Socket socket = ssocket.accept();

			// AXL[t@C̑M
			BufferedOutputStream output = new BufferedOutputStream(socket.getOutputStream());
			output.write(src,0,src.length);

			// Xg[
			output.close();
		}
		catch(IOException io_ex)
		{
			io_ex.printStackTrace();
			ErrorMessage("File IO Error");
		}

		if(!checkReturnCode(TRANSFER))
			ErrorMessage("File Transfer Failed");
	}

	/**
		擾/MItZbg̎w
		@param	擾/MItZbgA<br>
		0Zbg邱ƂɂāAItZbg
	*/
	public void setOffset(int offset)
	{
		this.offset = offset;
	}

	/**
		RESTR}hsɂA擾/MItZbg̎w
	*/
	private void setOffsetCommand()
	{
		sendData("REST " + offset);
		if(!checkReturnCode(FILE_OTHER))
			ErrorMessage("Set Offset Failed.");
	}

	/**
		t@CXg̎擾
		@return	t@CXg(ls -alɂ)
	*/
	public String[] getFileList()
	{
		// t@CXg
		Vector file_list = new Vector();

		// AXL[[hɃZbg
		selectType(TYPE_ASCII);

		// NCAg̏
		openClientPort();

		sendData("NLST -alL");
		if(!checkReturnCode(OPEN))
			ErrorMessage("Get File List Failed");

		try
		{
			// T[o̐ڑ̎󂯕t
			Socket socket = ssocket.accept();

			// Xg[̍쐬
			BufferedReader reader = new BufferedReader(new InputStreamReader(socket.getInputStream()));

			// Xg̎擾
			String buf;
			while((buf = reader.readLine()) != null)
			{
				file_list.add(buf);
			}

			// Xg[
			reader.close();

			if(!checkReturnCode(TRANSFER))
				ErrorMessage("File Transfer Failed");
		}
		catch(IOException io_ex)
		{
			io_ex.printStackTrace();
			ErrorMessage("File IO Error");
		}

		String file_list_array[] = new String[file_list.size()];
		file_list.toArray(file_list_array);
		return(file_list_array);
	}

	/**
		t@C폜
		@param	file_name	폜t@C
	*/
	public void removeFile(String file_name)
	{
		String current_directory = getCurrentDirectory();
		sendData("DELE " + current_directory + server_file_separator + file_name);
		if(!checkReturnCode(FILE))
			ErrorMessage("Can't delete " + file_name);
	}

	/**
		t@Cl[
		@param	from_file_name	ύXÕt@C
		@param	to_file_name	ύX̃t@C
	*/
	public void renameFile(String from_file_name, String to_file_name)
	{
		sendData("RNFR " + from_file_name);
		if(!checkReturnCode(FILE_OTHER))
			ErrorMessage(from_file_name + " is'nt exist");

		sendData("RNTO " + to_file_name);
		if(!checkReturnCode(FILE))
			ErrorMessage("Rename is Failed");
	}

	/**
		t@Cp[~bVύX
		@param	file_name	Ώۃt@C
		@param	permission	p[~bV̐l\(755644Ȃ)
	*/
	public void changePermission(String file_name, String permission)
	{
		sendData("SITE CHMOD " + permission + " " + file_name);
		if(!checkReturnCode(COMMAND))
			ErrorMessage("Change Permission Failed");
	}

	/**
		JgfBNg擾
		@return	JgfBNg
	*/
	public String getCurrentDirectory()
	{
		sendData("PWD");
		String server_message = recvData();
		if(!checkReturnCode(server_message,DIRECTORY))
			ErrorMessage("Can't get Current Directory");

		int begin_index = server_message.indexOf("\"");
		int end_index  = server_message.indexOf("\"",begin_index + 1);
		String current_directory = server_message.substring(begin_index + 1,end_index);

		return(current_directory);
	}

	/**
		fBNgړ
		@param	directory_name	ړ̃fBNg
	*/
	public void changeDirectory(String directory_name)
	{
		sendData("CWD " + directory_name);
		if(!checkReturnCode(FILE))
			ErrorMessage("Can't move to " + directory_name);
	}

	/**
		fBNg쐬
		@param	directory_name	쐬fBNg
	*/
	public void makeDirectory(String directory_name)
	{
		sendData("MKD " + directory_name);
		if(!checkReturnCode(DIRECTORY))
			ErrorMessage("Can't make directory");
	}

	/**
		fBNg폜
		@param	directory_name	폜fBNg
	*/
	public void removeDirectory(String directory_name)
	{
		String current_directory = getCurrentDirectory();
		sendData("RMD " + current_directory + server_file_separator + directory_name);
		if(!checkReturnCode(FILE))
			ErrorMessage("Can't remove directory");
	}

	/**
		VXeR}h̔s
		@return	VXe
	*/
	private String sendSystemCommand()
	{
		sendData("SYST");
		String server_message = recvData();
		if(!checkReturnCode(server_message,NAME))
			ErrorMessage("System Command Failed");

		return(server_message);
	}

	/**
		wvR}h̔s
		@return wv(gp\ȃR}hꗗ)
	*/
	public String help()
	{
		setSoTimeout(1000);

		sendData("HELP");
		String server_message = recvData();
		if(!checkReturnCode(server_message, HELP))
			ErrorMessage("Can't get help");

		String buf;
		StringBuffer buffer = new StringBuffer();
		do
		{
			buf = recvData();
			buffer.append(buf);
		} while(buf != null);

		setSoTimeout(0);

		return(buffer.toString());
	}

	/**
		FTPڑ̏I
	*/
	public void quit()
	{
		sendData("QUIT");
		if(!checkReturnCode(QUIT))
			ErrorMessage("Quit Failed");
	}

	/**
		wR}hAT[oɑΉĂ邩ǂ𒲂ׂ
		@param	command_name	R}h
		@return	T[oΉĂtrueAłȂfalseԂ
	*/
	public boolean isCommandAvailable(String command_name)
	{
		String help_message = help();
		if(help_message.indexOf(command_name) == -1)
			return(false);
		return(true);
	}

	/**
		G[bZ[W\[`
		@param	message	\G[bZ[W
	*/
	private void ErrorMessage(String message)
	{
		System.err.println(message);
		System.exit(-1);
	}

	/**
		sɓn鉞
		@param	expect_code	҂R[hio[(3)
	*/
	private void checkReturnCodeLoop(int expect_code)
	{
		setSoTimeout(1000);

		String buf;
		while(true)
		{
			buf = recvData();
			if(buf == null)
			{
				break;
			}
			else
			{
				if(!checkReturnCode(buf, expect_code))
					ErrorMessage("Invalid Server Response");
			}
		}

		setSoTimeout(0);
	}
}